"""
@author:  Patrice Neff
@since:   2003-05-08
@license: GNU General Public License
@version: $Revision: 1.15 $

Classes to handle different bookmark file formats.
"""

import sys
import xml.sax
from Node import NodeObserver, Folder, Bookmark, Separator, Metadata
from Xml import att_enc, chr_enc

class FormatObserver:
    """Abstract base class for receiving observer events.  The class
    functions are called from the class FormatSubject.

    @see: L{FormatSubject}
    @see: U{Observer Pattern<http://c2.com/cgi/wiki?ObserverPattern>}
    @since: 2003-05-08
    """
    def created(self, node):
        """Abstract function. Called by FormatSubject when a node was
        created.

        @param node: New node
        @type node: L{Node}
        """
        pass # implemented in concrete class
    def changed(self, node):
        """Abstract function. Called by FormatSubject when a node was
        modified.

        @param node: Modified node
        @type node: L{Node}
        """
        pass # implemented in concrete class
    def moved(self, node):
        """Abstract function. Called by FormatSubject when a node was
        moved.

        @param node: Modified node
        @type node: L{Node}
        """
        pass # implemented in concrete class
    def deleted(self, node):
        """Abstract function. Called by FormatSubject when a node was
        deleted.

        @param node: Deleted node
        @type node: L{Node}
        """
        pass # implemented in concrete class

class FormatSubject:
    """Abstract class which notifies subscribers about changes a
    format.

    @see: U{Observer Pattern<http://c2.com/cgi/wiki?ObserverPattern>}
    @since: 2003-05-08
    """
    def __init__(self):
        """Constructor."""
        self.__observers = []

    def subscribe(self, observer):
        """Add the observer to the list of observers so that it gets
        notified on changes.

        @param observer: Observer to add
        @type observer: L{FormatObserver}
        """
        if not observer in self.__observers:
            self.__observers.append(observer)

    def unsubscribe(self, observer):
        """Remove the observer from the list of observers. It won't
        receive any events anymore.

        @param observer: Observer to remove
        @type observer: L{FormatObserver}
        """
        try:
            self.__observers.remove(observer)
        except ValueError:
            pass

    def _notifyCreated(self, node):
        """Notify all observers that a new node was added.

        @param node: New node
        @type node: L{Node}
        """
        for observer in self.__observers:
            observer.created(node)

    def _notifyChanged(self, node):
        """Notify all observers that a node was modified.

        @param node: Modified node
        @type node: L{Node}
        @bug: This event is not executed from Format class yet
        """
        for observer in self.__observers:
            observer.changed(node)

    def _notifyMoved(self, node):
        """Notify all observers that a node was moved.

        @param node: Moved node
        @type node: L{Node}
        @todo: Also execute when a node is changed, not only when
               moved.
        """
        for observer in self.__observers:
            observer.moved(node)

    def _notifyDeleted(self, node):
        """Notify all observers that a node was deleted.

        @param node: Deleted node
        @type node: L{Node}
        """
        for observer in self.__observers:
            observer.deleted(node)

class Format(FormatSubject):
    """Abstract base class for all the format classes.

    @since: 2003-05-08
    """
    def __init__(self, filename = ''):
        """Constructor. Load the database from the file in memory.
        
        @param filename: File to open. If the filename is empty, a new
                         database is created.
        @type filename: string
        """
        FormatSubject.__init__(self)
    def open(self, filename):
        """Close the current database and open an other one.
        
        @param filename: File to open. If the filename is empty, a new
                         database is created.
        @type filename: string
        @return: False on error, True otherwise
        @rtype: Boolean
        """
    def save(self, filename = ''):
        """Save the database from memory to a file.
        
        @param filename: Filename to save as. If no filename is
                         given, the database is saved to the file
                         where is was opened from.
        
        @type filename: string
        @return: False on error, True otherwise
        @rtype: Boolean
        """
        pass # implemented in concrete class
    def getFilename(self):
        """Return the file name of the currently opened file.
        
        @return: File name
        @rtype: String
        """
        pass # implemented in concrete class
    def getFirst(self):
        """Get the root folder of the database.
        
        @return: Root folder
        @rtype: L{Folder}
        """
        pass # implemented in concrete class
    def search(self, query):
        """Search for a node in the whole database and return all
        matches.
        
        @param query: Term to look for
        @type query: string
        @return: All matched nodes
        @rtype: List of L{Node}
        """
        pass # implemented in concrete class
    def add(self, node):
        """Add the node to the database. The nodes parent must already
        be set to the folder where the node should be stored in. If no
        parent is set, it is saved into the root folder.

        @param node: Node to add to the database
        @type node: L{Node}
        @return: False on error, True otherwise
        @rtype: Boolean
        """
        pass # implemented in concrete class
    def remove(self, node):
        """Remove the node from the database. Also handles relinking
        the nodes (first child and next sibling). The node object is
        not destroyed and references to it are still valid after
        calling this method.

        @param node: Node to remove from the database
        @type node: L{Node}
        @return: False on error, True otherwise
        @rtype: Boolean
        """
        pass # implemented in concrete class
    def move(self, node, newParent):
        """Move the node to a new parent. This automatically handles
        relinking the nodes and notifies all observers about the
        change.
        @attention: It would be possible to relink nodes without
                    calling this method. But this method should still
                    be used, mainly for notifying the
                    observers. Additionally, implementing classes are
                    allowed to depend on this fact if it is necessary
                    for implementing some of the algorithms.

        @param node: Node to add to the database
        @type node: L{Node}
        @param newParent: Folder where node is moved into
        @type newParent: L{Folder}
        @return: False on error, True otherwise
        @rtype: Boolean
        """
        pass # implemented in concrete class

class XBEL(Format,NodeObserver):
    """Format class for the XBEL data format.

    @see: U{The XML Bookmark Exchange Language (XBEL)
            <http://pyxml.sourceforge.net/topics/xbel/>}
    @since: 2003-05-08
    @todo: Handle alias
    @todo: The XBEL specification allows any content in the metadata
           elements. We should take care of that and not exclusively
           allow key/value pairs. It must be possible to store other
           data as well in order to save it again.
    """
    class XBELReader(xml.sax.handler.ContentHandler):
        """SAX ContentHandler class to read XBEL data.

        @since: 2004-02-13
        """
        def __init__(self, xbel):
            """Constructor. Assign the object to a XBEL format object.

            @param xbel: XBEL format class reference
            @type xbel: L{XBEL}
            """
            self.__xbel = xbel     # xbel document where nodes are added to
            self.__folders = []    # currently processed folders
            self.__node = None     # currently processed node
            self.__metadata_owner = '' # owner of current metadata
            self.__chrs = u''

        def startElement(self, name, attrs):
            """Signals the start of an element in non-namespace mode.
            
            The name parameter contains the raw XML 1.0 name of the
            element type as a string and the attrs parameter holds an
            instance of the Attributes class containing the attributes
            of the element.

            @param name: element name
            @type name: Unicode string
            @param attrs: Element attributes. Values can be retrieved
                          using C{attrs.get(key, default_value)}.
            @type attrs: Unicode string
            """
            if name == 'xbel':
                self.__node = Folder(None, '', u'Root')
                self.__folders.append(self.__node)
                self.__xbel._setRoot(self.__node)
            elif name == 'folder':
                self.__node = Folder(self.__getCurrentParent())
                self.__folders.append(self.__node)
                self.__node.setId(attrs.get('id', ''))
                self.__node.setAdded(attrs.get('added', ''))
                if attrs.get('folded',"") == "yes":
                    self.__node.setFolded(True)
                self.__xbel.add(self.__folders[-1], False)
            elif name == 'bookmark':
                if self.__folders:
                    self.__node = Bookmark(self.__getCurrentParent(), attrs.get('href', ''))
                    self.__node.setId(attrs.get('id', ''))
                    self.__node.setAdded(attrs.get('added', ''))
                    self.__node.setModified(attrs.get('modified', ''))
                    self.__node.setVisited(attrs.get('visited', ''))
                    self.__xbel.add(self.__node, False)
            elif name == 'separator':
                if self.__folders:
                    sep = Separator(self.__getCurrentParent())
                    self.__xbel.add(sep, False)
            elif name == 'metadata':
                self.__metadata_owner = attrs.get('owner', '')
                self.__metadata = Metadata(self.__metadata_owner)
            self.__chrs = ''

        def endElement(self, name):
            """Signals the end of an element in non-namespace mode.

            The name parameter contains the name of the element type,
            just as with the startElement event.

            @param name: element name
            @type name: Unicode string
            """
            if self.__node is None:
                return
            if name == 'folder':
                if self.__folders:
                    self.__node = self.__folders.pop()
            elif name == 'bookmark':
                if self.__folders:
                    self.__node = self.__folders[-1]
            elif name == 'title':
                self.__node.setTitle(self.__chrs)
            elif name == 'desc':
                self.__node.setDesc(self.__chrs)
            elif name == 'metadata':
                self.__node.addMetadata(self.__metadata)
                self.__metadata_owner = ''
            elif self.__metadata_owner != '':
                # in metadata, every tag is one metadata entry
                self.__metadata.setValue(name, self.__chrs)

        def characters (self, content):
            """Receive notification of character data.
            
            The Parser will call this method to report each chunk of
            character data. SAX parsers may return all contiguous
            character data in a single chunk, or they may split it
            into several chunks; however, all of the characters in any
            single event must come from the same external entity so
            that the Locator provides useful information.

            @param content: Character data
            @type content: Unicode string
            """
            self.__chrs += content

        def __getCurrentParent(self):
            """This returns the folder object that should be used as
            parent. This is usually the latest item in the __folders
            stack. If there is non, the root object is used (if
            necessary this root object created).

            @return: Parent folder
            @rtype: L{Folder}
            """
            parent = None
            if self.__folders:
                parent = self.__folders[-1]
            else:
                parent = self.__xbel.getFirst()
            if parent is None:
                self.__xbel._setRoot()
            return parent


    def __init__(self, filename = ''):
        Format.__init__(self, filename)
        self.open(filename)
        if self.getFilename() == '':
            # file not successfully opened
            self._setRoot()

    def open(self, filename):
        """Close the current database and open an other one.
        
        @param filename: File to open. If the filename is empty, a new
                         database is created.
        @type filename: string
        @return: False on error, True otherwise
        @rtype: Boolean
        """
        self.__filename = ''
        self.__root = None
        try:
            parser = xml.sax.parse(filename, XBEL.XBELReader(self))
        except IOError,(errno, strerror):
	    pass
        except ValueError,(strerror):
            if str(strerror).find('unknown url type:') != -1:
                pass
            else:
                print "Unexpected error:", sys.exc_info()[0]
                raise
        except:
            print "Unexpected error:", sys.exc_info()[0]
            raise
        else:
            self.__filename = filename

    def parse(self, data):
        """Parse the data string and store the XBEL structure in this
        object as if a file had been opened.
        
        @param data: XBEL Data to parse
        @type data: string
        @return: False on error, True otherwise
        @rtype: Boolean
        """
        self.__root = None
        try:
            if data.strip().find("<bookmark") == 0 or data.strip().find("<folder") == 0:
                # if data comes directly without any xbel element, store data inside
                # xbel element.
                data = '<xbel version="1.0">' + data + "</xbel>"
            parser = xml.sax.parseString(data, XBEL.XBELReader(self))
        except:
            print "Unexpected error:", sys.exc_info()[0]
            raise

    def save(self, filename = ''):
        if filename == '':
            filename = self.__filename
        if filename == '':
            return False

        file = open(filename, 'w')

        root = self.getFirst()
        file.write('<?xml version="1.0"?>\n')
	root.writeXbel(file)
        file.close()
        self.__filename = filename
        return True

    def getFirst(self):
        return self.__root

    def search(self, query):
        matches = []

        item = self.getFirst()
        while item is not None:
            if item.match(query):
                matches.append(item)
            try:
                nextitem = item.getFirstChild()
            except:
                nextitem = None
            if nextitem is None:
                nextitem = item.getNextSibling()
                if nextitem is None:
                    nextitem = item.getParent()
                    if nextitem is None:
                        return matches
                    while nextitem.getNextSibling() is None:
                        nextitem = nextitem.getParent()
                        if nextitem is None:
                            return matches
                    nextitem = nextitem.getNextSibling()
                    
            item = nextitem
        return matches

    def _setRoot(self, root = None):
        """Set the root node of the database.

        @param root: New root node
        @type root: L{Folder}
        """
        if root is None:
            root = Folder(None, '', u'Root')
        root.setParent(None)
        root.setNextSibling(None)
        if self.__root is not None:
            self.__root.unsubscribe(self)
        self.__root = root
        root.subscribe(self)
    
    def add(self, node, notify = True):
        if node.getParent() is None:
            # can't replace root
            node.setParent(self.__root)

        # link node into tree
        sibling = node.getParent().getFirstChild()
        if sibling is None:
            node.getParent().setFirstChild(node)
        else:
            while sibling.getNextSibling() is not None:
                sibling = sibling.getNextSibling()
            sibling.setNextSibling(node)
            node.setNextSibling(None)
        node.subscribe(self)
        if notify:
            self._notifyCreated(node)

    def remove(self, node, notify = True):
        if node == self.__root:
            # deleting root is now allowed
            return False
        node.unsubscribe(self)

        # unlink node from tree
        tmp = node.getParent().getFirstChild()
        if tmp == node:
            node.getParent().setFirstChild(node.getNextSibling())
        else:
            while tmp.getNextSibling() != node:
                tmp = tmp.getNextSibling()
            tmp.setNextSibling(node.getNextSibling())
        node.setParent(None)
        node.setNextSibling(None)

        if notify:
            self._notifyDeleted(node)
        return True
                
    def move(self, node, newParent):
        if node == self.__root:
            # moving root is not allowed
            return False
        if self.remove(node, False):
            node.setParent(newParent)
            self.add(node, False)
            self._notifyMoved(node)
        else:
            return False
        return True
    
    def getFilename(self):
        return self.__filename

    def _dumpTree(self):
        """Print the database in a nicely formatted way to the command
        line.
        
        @depreciated: This function is intended solely for debugging
                      purposes and will be removed from the final
                      version.
        """
        indent = 0
        print "\n===BEGIN DUMP==="
        item = self.getFirst()
        while item is not None:
            print indent*' '+item.getTitle()
            try:
                nextitem = item.getFirstChild()
            except:
                nextitem = None
            else:
                if nextitem is not None:
                    indent+=2
            if nextitem is None:
                nextitem = item.getNextSibling()
                if nextitem is None:
                    nextitem = item.getParent()
                    indent-=2
                    if nextitem is None:
                        print "===END DUMP===\n"
                        return
                    while nextitem.getNextSibling() is None:
                        nextitem = nextitem.getParent()
                        indent-=2
                        if nextitem is None:
                            print "===END DUMP===\n"
                            return
                    nextitem = nextitem.getNextSibling()
            item = nextitem
        print "===END DUMP===\n"

    def changed(self, node):
        """A node was changed. Implements the NodeObserver interface.

        @param node: Changed node
        @type node: L{Node}
        """
        self._notifyChanged(node)
