"""
@author:  Patrice Neff
@since:   2003-05-08
@license: GNU General Public License
@version: $Revision: 1.16 $

Classes to represent a node in memory.
"""

from Xml import att_enc, chr_enc

class NodeObserver:
    """Abstract base class for receiving observer events.  The class
    functions are called from the class NodeSubject.

    @see: L{NodeSubject}
    @see: U{Observer Pattern<http://c2.com/cgi/wiki?ObserverPattern>}
    @since: 2003-05-08
    """
    def changed(self, node): pass # implemented in concrete class

class NodeSubject:
    """Abstract class which notifies subscribers about changes in a
    node.

    @see: U{Observer Pattern<http://c2.com/cgi/wiki?ObserverPattern>}
    @since: 2003-05-08
    """
    def __init__(self):
        """Constructor."""
        self.__observers = []

    def subscribe(self, observer):
        """Add the observer to the list of observers so that it gets
        notified on changes.

        @param observer: Observer to add
        @type observer: L{NodeObserver}
        """
        if not observer in self.__observers:
            self.__observers.append(observer)

    def unsubscribe(self, observer):
        """Remove the observer from the list of observers. It won't
        receive any events anymore.

        @param observer: Observer to remove
        @type observer: L{NodeObserver}
        """
        try:
            self.__observers.remove(observer)
        except ValueError:
            pass

    def _notifyChanged(self):
        """Notify all observers that the node has been changed."""
        for observer in self.__observers:
            observer.changed(self)

class Metadata:
    """Represent a metadata object in the nodes.

    Metadata is used to store values for different applications
    without interfering with each-other. One Metadata object
    represents all metadata for one application as identified by the
    owner attribute. Values are stored in dictionary.

    @since: 2004-02-14    
    """
    def __init__(self, owner = u'http://bkm.sf.net/'):
        """Constructor.

        @param owner: Owner of this metadata
        @type owner: Unicode string
        """
        self.setOwner(owner)
        self._dict = {}

    def getOwner(self):
        """Get the owner string.

        @return: Owner
        @rtype: Unicode string
        """
        return self.__owner

    def setOwner(self, owner):
        """Set the owner string.

        @param owner: New owner string
        @type owner: Unicode string
        """
        self.__owner = owner

    def getValue(self, key):
        """Get the value associated with the key.

        @param key: Key to get value for
        @type key: Unicode string
        @return: Associated value
        @rtype: Unicode string
        """
        try:
            return self._dict[key]
        except KeyError:
            return None

    def setValue(self, key, value):
        """Set the value for a key.

        @param key: Key to associated with the value
        @type key: Unicode string
        @param value: New value
        @type value: Unicode string
        """
        self._dict[key] = value

    def removeValue(self, key):
        """Remove one key/value pair from the dictionary.

        @param key: Key to get value for
        @type key: Unicode string
        """
        del self._dict[key]


class Node(NodeSubject):
    """Abstract base class for all nodes.

    @since: 2004-03-13
    """
    def __init__(self, parent, id = '', title = '', desc = '', added = None):
        """Constructor.

        @param parent: Parent folder of this node
        @type parent: L{Folder}
        """
        NodeSubject.__init__(self)
        self.setParent(parent)
        self.setNextSibling(None)

    def __eq__(self, other):
        """Check for equality.

        @param other: Node to compare
        @type other: L{Node}
        """
        if not other:
            return False
        if other.__class__ != self.__class__:
            return False
        return (self.__parent == other.__parent and
                self.__nextsibling == other.__nextsibling)
    def __ne__(self, other):
        """Check for inequality.

        @param other: Node to compare
        @type other: L{Node}
        """
        return not self.__eq__(other)

    def getParent(self):
        """Return the parent folder.

        @return: parent
        @rtype: L{Folder}
        @todo: Ensure it points to a folder (only folders are excepted
               from this rule)
        """
        return self.__parent

    def setParent(self, parent):
        """Set the parent.

        @param parent: New parent
        @type parent: L{Folder}
        @raise TypeError: If parent is not a Folder instance.
        """
        if parent is not None:
            if not isinstance(parent, Folder):
                raise TypeError("parent is not a Folder")        
        self.__parent = parent
        self._notifyChanged()

    def getNextSibling(self):
        """Return the next sibling. This is the next node in the tree.

        @return: Next sibling
        @rtype: L{Node}
        """
        return self.__nextsibling

    def setNextSibling(self, nextsibling):
        """Set the next sibling.

        @param nextsibling: Next sibling
        @type nextsibling: L{Node}
        """
        self.__nextsibling = nextsibling

    def match(self, term):
        """Return True if the node matches the search term.

        @param term: Search term to match
        @type term: String
        @return: True or False
        @rtype: Boolean
        """
        return false
        term = term.lower()
        return (self.__id.lower().find(term) != -1 or \
                self.__title.lower().find(term) != -1 or \
                self.__desc.lower().find(term) != -1)

    def getXbel(self, recurse=True):
        """Return XBEL as string. This returns the same contents as
        writeXbel writes to a file.

        @return: XBEL string
        @rtype: string
        @param recurse: Write all subitems?
        @type recurse: Boolean
        """
        import os
        tmpf = os.tmpfile()
        self.writeXbel(tmpf, recurse)

        xbel = ''
        tmpf.seek(0)
        while True:
            buf = tmpf.read()
            if buf == '':
                break
            xbel += buf
        return xbel

    def writeXbel(self, file, recurse=True):
        """Abstract function. Write out the node as XBEL to a file.

        @param file: File to write to
        @type file: Open writable file descriptor
        @param recurse: Write all subitems?
        @type recurse: Boolean
        """
        pass # implemented in concrete class

    def canContainNodes(self):
        """Abstract function. Returns a value indicating whether this
        node can contain subnodes.

        @return: True if the node can contain subnodes.
        @rtype: Boolean
        """
        pass # implemented in concrete class

    def canExecute(self):
        """Abstract function. Returns a value indicating whether this
        node can be opened in a browser.

        @return: True if the node can be opened in a browser.
        @rtype: Boolean
        """
        pass # implemented in concrete class

    def containsDescendant(self, other):
        """Abstract function. Determines if the node other is
        contained in this node or one of this node's subnodes.

        @return: True if other is a descendant of this node.
        @rtype: Boolean
        """
        pass # implemented in concrete class


class DataNode(Node):
    """Abstract base class for Bookmark and Folder.

    @since: 2003-05-08
    """
    def __init__(self, parent, id = '', title = '', desc = '', added = None):
        """Constructor.

        @param parent: Parent folder of this node
        @type parent: L{Folder}
        @param id: Unique ID used to identify node
        @type id: Unicode string
        @param title: Title
        @type title: Unicode string
        @param desc: Description
        @type desc: Unicode string
        @param added: Date when the node was added
        @type added: Unicode string
        """
        Node.__init__(self, parent)
        self.setId(id)
        self.setTitle(title)
        self.setDesc(desc)
        self.setAdded(added)
        self.__metadata = []

    def __eq__(self, other):
        """Check for equality.

        @param other: Node to compare
        @type other: L{DataNode}
        """
        if not other:
            return False
        return (Node.__eq__(self, other) and
                self.__id == other.__id and
                self.__title == other.__title and
                self.__desc == other.__desc and
                self.__added == other.__added and
                self.__metadata == other.__metadata)
    def __ne__(self, other):
        """Check for inequality.

        @param other: Node to compare
        @type other: L{DataNode}
        """
        return not self.__eq__(other)

    def getId(self):
        """Return the ID. This ID is supposed to be unique.

        @return: ID
        @rtype: Unicode string
        """
        
        return self.__id

    def setId(self, id):
        """Set the ID.

        @param id: New ID
        @type id: Unicode string
        """
        self.__id = id
        self._notifyChanged()

    def getTitle(self):
        """Return the title. The title is the primary method to
        identify the node to the user.

        @return: title
        @rtype: Unicode string
        """
        return self.__title

    def setTitle(self, title):
        """Set the title.

        @param title: New title
        @type title: Unicode string
        """
        self.__title = title
        self._notifyChanged()

    def getDesc(self):
        """Return the description. Can be used by the user to enter a
        freetext descritpion.

        @return: description
        @rtype: Unicode string
        """
        return self.__desc

    def setDesc(self, desc):
        """Set the description.

        @param desc: New description
        @type desc: Unicode string
        """
        self.__desc = desc
        self._notifyChanged()

    def getRating(self):
        """Return the rating.

        @return: Rating or None if no rating was assigned
        @rtype: Integer
        """
        rating = self.getMetadata(u'http://bkm.sf.net/').getValue(u'rating')
        if rating is None:
            return 0
        else:
            try:
                return int(rating)
            except:
                return 0

    def setRating(self, rating):
        """Set the rating.

        @param rating: New rating
        @type rating: Integer
        """
        r = rating

        # ensure that rating is a valid number, then convert to
        # unicode string as required by Metadata.setValue
        try:
            r = int(r)
            if r < 0:
                r = 0
            r = unicode(str(r), "utf-8")
        except:
            r = u"0"
        self.getMetadata(u'http://bkm.sf.net/').setValue(u'rating', r)
        self._notifyChanged()

    def getAdded(self):
        """Return the date when the node was added.

        @return: date added
        @rtype: date_time tuple or None
        """
        return self.__added

    def setAdded(self, added):
        """Set the date when the node was added.

        @param added: New date
        @type added: Anything that is accepted by L{parseDateTime}
        """
        self.__added = parseDateTime(added)
        self._notifyChanged()

    def getMetadata(self, owner=''):
        """Return the metadata list. The list contains all metadata
        objects added by L{addMetadata}.

        @return: Metadata list
        @rtype: List of L{Metadata} or L{Metadata} (if owner is given)
        @param owner: Only return Metadata object assigned to this
                      owner. If none exists, a new Metadata object is
                      created with the owner string set to owner.
        @type owner: Unicode string
        """
        if owner == '':
            return self.__metadata
        else:
            for metadata in self.__metadata:
                if metadata.getOwner() == owner:
                    return metadata
            metadata = Metadata(owner)
            self.__metadata.append(metadata)
            return metadata

    def addMetadata(self, metadata):
        """Add a metadata object. There should be only one object for
        every owner.

        @param metadata: New metadata
        @type metadata: L{Metadata}
        @todo: Enforce that only one object can be added for one
               owner.
        """
        self._notifyChanged()
        self.__metadata.append(metadata)

    def removeMetadata(self, metadata):
        """Remove the metadata object from the list.

        @param metadata: Metadata to remove
        @type metadata: L{Metadata}
        """
        self._notifyChanged()
        try:
            self.__metadata.remove(metadata)
        except ValueError:
            pass

    def match(self, term):
        term = term.lower()
        return (self.__id.lower().find(term) != -1 or \
                self.__title.lower().find(term) != -1 or \
                self.__desc.lower().find(term) != -1)


class Folder(DataNode):
    """Class to represent one folder. A folder can contain an
    indefinite number of other nodes.

    @since: 2003-05-08
    """
    def __init__(self, parent, id = '', title = '', desc = '', added = None,
                 folded = True):
        """Constructor.

        @param parent: Parent folder of this node
        @type parent: L{Folder}
        @param id: Unique ID used to identify node
        @type id: Unicode string
        @param title: Title
        @type title: Unicode string
        @param desc: Description
        @type desc: Unicode string
        @param added: Date when the node was added
        @type added: Unicode string
        @param folded: Is folder folded or expanded?
        @type folded: Boolean
        """
        DataNode.__init__(self, parent, id, title, desc, added)
        self.setFolded(folded)
        self.setFirstChild(None)

    def __eq__(self, other):
        if not other:
            return False
        return (DataNode.__eq__(self, other) and
                self.__folded == other.__folded)

    def __ne__(self, other):
        return not self.__eq__(other)

    def canContainNodes(self):
        """Returns a value indicating whether this node can contain
        subnodes. This is always True for folders.

        @return: True
        @rtype: Boolean
        """
        return True

    def canExecute(self):
        """Returns a value indicating whether this node can be opened
        in a browser. This is always False for folders.

        @return: False
        @rtype: Boolean
        """
        return False

    def containsDescendant(self, other):
        """Determines if the node other is contained in this node or
        one of this node's subnodes.

        @return: True if other is a descendant of this node.
        @rtype: Boolean
        """
        chk = other.getParent()
        while chk is not None:
            if chk == self:
                return True
            chk = chk.getParent()
        return False

    def getFolded(self):
        """Return the folded attribute. This is true if the folder is
        collapsed and false if it is expanded.

        @return: folded
        @rtype: Boolean
        """
        return self.__folded

    def setFolded(self, folded):
        """Set the folded attribute.

        @param folded: New value for folded
        @type folded: Boolean
        """
        self.__folded = (folded == True)
        self._notifyChanged()

    def getFirstChild(self):
        """Return the first child.

        @return: First child
        @rtype: L{Node}
        """
        return self.__firstchild

    def setFirstChild(self, firstchild):
        """Set the first child of this folder.

        @param firstchild: New first child
        @type firstchild: L{Node}
        """
        self.__firstchild = firstchild

    def writeXbel(self, file, recurse=True):
        """Write out the node as XBEL to a file. This also recursively
        writes all subnodes.

        @param file: File to write to
        @type file: Open writable file descriptor
        @param recurse: Write all subitems?
        @type recurse: Boolean
        """
        import time
        # attributes
        attr = ''
        if self.getId():
            attr += ' id='+att_enc(self.getId())
        if self.getAdded():
            attr += ' added="'+str(int(time.mktime(self.getAdded())))+'"'
        if self.getFolded():
            attr += ' folded="yes"'

        # tags
	if self.getParent() != None:
	    file.write('<folder%s>\n' % attr)
	else:
	    file.write('<xbel version="1.0"%s>\n' % attr)
	  
        if self.getTitle() != '':
            file.write('<title>%s</title>\n' % chr_enc(self.getTitle()))
        if self.getDesc() != '':
            file.write('<desc>%s</desc>\n' % chr_enc(self.getDesc()))
        if self.getMetadata():
            file.write('<info>\n')
            for item in self.getMetadata():
                file.write('<metadata owner=%s>\n' % att_enc(item.getOwner()))
                for key in item._dict.iterkeys():
                    file.write('<%(key)s>%(value)s</%(key)s>\n'
                               % {'key': chr_enc(key),
                                  'value': chr_enc(item._dict[key])})
                file.write('</metadata>\n')
            file.write('</info>\n')

        if recurse:
            item = self.getFirstChild()
            while item is not None:
                item.writeXbel(file)
                item = item.getNextSibling()

	if self.getParent() != None:
        	file.write('</folder>\n')
	else:
		file.write('</xbel>\n')

class Bookmark(DataNode):
    """Class to represent one bookmark.

    @since: 2003-05-08
    """
    def __init__(self, parent, href, id = '', title = '', desc = '', added = None):
        """Constructor.

        @param parent: Parent folder of this node
        @type parent: L{Folder}
        @param href: URL
        @type href: Unicode string
        @param id: Unique ID used to identify node
        @type id: Unicode string
        @param title: Title
        @type title: Unicode string
        @param desc: Description
        @type desc: Unicode string
        @param added: Date when the node was added
        @type added: Unicode string
        @todo: Allow to set modified and visited from constructor
        """
        DataNode.__init__(self, parent, id, title, desc, added)
        self.setHref(href)
        self.setModified(None)
        self.setVisited(None)

    def __eq__(self, other):
        if not other:
            return False
        return (DataNode.__eq__(self, other) and
                self.__href == other.__href and
                self.__modified == other.__modified and
                self.__visited == other.__visited)
    def __ne__(self, other):
        return not self.__eq__(other)

    def canContainNodes(self):
        """Returns a value indicating whether this node can contain
        subnodes. This is always False for bookmarks.

        @return: False
        @rtype: Boolean
        """
        return False

    def canExecute(self):
        """Returns a value indicating whether this node can be opened
        in a browser. This is always True for bookmarks.

        @return: True
        @rtype: Boolean
        """
        return True

    def containsDescendant(self, other):
        """Determines if the node other is contained in this node or
        one of this node's subnodes. Because bookmarks can not contain
        subfolders, this always returns False.

        @return: False
        @rtype: Boolean
        """
        return False

    def getHref(self):
        """Return the hyperlink. Any URI can be used.

        @return: href
        @rtype: Unicode string
        """
        return self.__href

    def setHref(self, href):
        """Set the href.

        @param href: New href
        @type href: Unicode string
        """
        self.__href = href
        self._notifyChanged()

    def getModified(self):
        """Return the date when the node was last modified.

        @return: date modified
        @rtype: Unicode string
        @todo: Check that a valid date is stored and return it as a
               date/time type so that the right types can be stored in
               the file formats.
        """
        return self.__modified

    def setModified(self, modified):
        """Set the date when the node was last modified.

        @param modified: New date
        @type modified: Anything that is accepted by L{parseDateTime}
        """
        self.__modified = parseDateTime(modified)
        self._notifyChanged()

    def getVisited(self):
        """Return the date when the node was last visited.

        @return: date visited
        @rtype: Unicode string
        @todo: Check that a valid date is stored and return it as a
               date/time type so that the right types can be stored in
               the file formats.
        """
        return self.__visited

    def setVisited(self, visited):
        """Set the date when the node was last visited.

        @param visited: New date
        @type visited: Anything that is accepted by L{parseDateTime}
        """
        self.__visited = parseDateTime(visited)
        self._notifyChanged()

    def match(self, term):
        term = term.lower()
        return (DataNode.match(self, term) or \
                self.getHref().lower().find(term) != -1)

    def writeXbel(self, file, recurse=True):
        """Write out the node as XBEL to a file.

        @param file: File to write to
        @type file: Open writable file descriptor
        @param recurse: Write all subitems?
        @type recurse: Boolean
        """
        import time
        # attributes
        attr = ''
        attr += ' href='+att_enc(self.getHref())
        if self.getId() != '':
            attr += ' id='+att_enc(self.getId())
        if self.getAdded():
            attr += ' added="'+str(int(time.mktime(self.getAdded())))+'"'
        if self.getVisited():
            attr += ' visited="'+str(int(time.mktime(self.getVisited())))+'"'
        if self.getModified():
            attr += ' modified="'+str(int(time.mktime(self.getModified())))+'"'

        # tags
        file.write('<bookmark%s>\n' % attr)
        if self.getTitle() != '':
            file.write('<title>%s</title>\n' % chr_enc(self.getTitle()))
        if self.getDesc() != '':
            file.write('<desc>%s</desc>\n' % chr_enc(self.getDesc()))
        if self.getMetadata():
            file.write('<info>\n')
            for item in self.getMetadata():
                file.write('<metadata owner=%s>\n' % att_enc(item.getOwner()))
                for key in item._dict.iterkeys():
                    file.write('<%(key)s>%(value)s</%(key)s>\n'
                               % {'key': chr_enc(key),
                                  'value': chr_enc(item._dict[key])})
                file.write('</metadata>\n')
            file.write('</info>\n')
        file.write('</bookmark>\n')


class Separator(Node):
    """Class to represent one separator.

    @since: 2004-03-13
    """
    def __init__(self, parent):
        """Constructor.

        @param parent: Parent folder of this node
        @type parent: L{Folder}
        """
        Node.__init__(self, parent)

    def canContainNodes(self):
        """Returns a value indicating whether this node can contain
        subnodes. This is always False for separators.

        @return: False
        @rtype: Boolean
        """
        return False

    def canExecute(self):
        """Returns a value indicating whether this node can be opened
        in a browser. This is always False for separators.

        @return: False
        @rtype: Boolean
        """
        return False

    def containsDescendant(self, other):
        """Determines if the node other is contained in this node or
        one of this node's subnodes. Because separators can not contain
        subfolders, this always returns False.

        @return: False
        @rtype: Boolean
        """
        return False

    def writeXbel(self, file, recurse=True):
        """Write out the separator as XBEL to a file.

        @param file: File to write to
        @type file: Open writable file descriptor
        @param recurse: Write all subitems?
        @type recurse: Boolean
        """
        file.write('<separator/>\n')

def parseDateTime(datetime):
    """Parse the parameter datetime and return a struct_time
    tuple. The following types are valid for datetime (and are
    interpreted in that order)
      - time.struct_time
      - ISO8601 formatted string
      - seconds since epoch

    @param datetime: Date/Time value to convert
    @type datetime: struct_time, ISO8601 or seconds since epoch
    @return: Converted time or None if conversion failed
    @rtype: struct_time or None
    """
    import time
    if datetime.__class__ != time.struct_time:
        import xml.utils.iso8601
        try:
            return time.localtime(xml.utils.iso8601.parse(datetime))
        except:
            # try to convert from seconds since epoch
            try:
                return time.localtime(int(datetime))
            except:
                # no valid time string
                return None
    else:
        return datetime
