# -*- coding: iso-8859-1 -*-
#
# Author:    Patrice Neff
# Purpose:   Unittests for the Data.Formats
#
# History:
#   - 2003-05-08  created, pn
#   - 2004-01-23  XBEL tests expaned, pn

import unittest
import os
from time import localtime
from Formats import *

class XBELTest(unittest.TestCase):
    def testNonExisting(self):
        file = XBEL('doesnotexist.xbel')
        assert file.getFilename() == ''
        assert file.getFirst() != None
	assert file.getFirst().getTitle() == 'Root'
	assert file.getFirst() == Folder(None, '', u'Root')
    
    def testFilename(self):
        file = XBEL('bookmarks.xbel')
        assert file.getFilename() == 'bookmarks.xbel', 'filename not set'
        file.open('test_does_not_exist.xbel')
        assert file.getFilename() == '', 'filename set although test_does_not_exist.xbel does not exist'
        file.open('bookmarks.xbel')
        assert file.getFilename() == 'bookmarks.xbel', 'filename not set'

    def testGetFirst(self):
        file = XBEL('bookmarks.xbel')
        node = file.getFirst()
        assert node, 'getFirst returned no item'
        assert node.getTitle() == 'PyUnit test bookmarks', 'getFirst returned wrong item'

    def testGetFirstChild(self):
        file = XBEL('bookmarks.xbel')
        node = file.getFirst()
        assert node, 'getFirst returned no node'
        node = node.getFirstChild()
        assert node, 'getFirstChild returned no node'
        assert node.getTitle() == 'Python', 'getFirstChild returned wrong node'
        assert node.getFolded() == True, 'getFirstChild returned wrong node'
        node = node.getFirstChild()
        assert node, 'getFirstChild returned no node'
        assert node.getTitle() == 'python.org', 'getFirstChild returned wrong node'
        assert node.getHref() == 'http://www.python.org', 'getFirstChild returned wrong node'
        assert node.getParent(), 'getFirstChild node has no parent'
        assert node.getParent().getTitle() == 'Python', 'getFirstChild node has wrong parent'

    def testGetNextSibling(self):
        file = XBEL('bookmarks.xbel')
        node = file.getFirst()
        assert node, 'getFirst returned no node'
        node = node.getFirstChild().getFirstChild()
        assert node, 'getFirstChild returned no node'
        node = node.getNextSibling()
        assert node, 'getNextSibling returned no node'
        assert node.getTitle() == 'pyunit', 'getNextSibling returned wrong node'
        assert node.getHref() == 'http://pyunit.sourceforge.net', 'getNextSibling returned wrong node'
        node = node.getNextSibling()
        assert node, 'getNextSibling returned no node'
        assert node.getTitle() == 'Python Subfolder', 'getNextSibling returned wrong node'
        assert node.getFolded() == True, 'getNextSibling returned wrong value for folded'
        node = node.getNextSibling()
        assert node == None, 'getNextSibling returned node but should not have'

    def testSearch(self):
        file = XBEL('bookmarks.xbel')
        nodes = file.search('python')
        assert len(nodes) == 4, 'Did not get correct number of search results'
        assert nodes[0].getTitle() == 'Python', 'Invalid search result returned'
        assert nodes[1].getTitle() == 'python.org', 'Invalid search result returned'
        assert nodes[1].getHref() == 'http://www.python.org', 'Invalid search result returned'
        assert nodes[2].getTitle() == 'pyunit', 'Invalid search result returned'
        assert nodes[3].getTitle() == 'Python Subfolder', 'Invalid search result returned'

        nodes = file.search('unit')
        assert len(nodes) == 2, 'Did not get correct number of search results'
        assert nodes[0].getTitle() == 'PyUnit test bookmarks'
        assert nodes[1].getTitle() == 'pyunit'
        assert nodes[1].getHref() == 'http://pyunit.sourceforge.net'
        assert nodes[1].getDesc() == 'Python unit testing.'

        nodes = file.search('gutelaune')
        assert len(nodes) == 0, 'Returned result for inexistient item'

    def testAdd(self):
        file = XBEL('bookmarks.xbel')
        nodes = file.search('python')
        # append an item to the 'Python Subfolder'
        newBookmark = Bookmark(nodes[3], u'http://www.patrice.ch/',
                               u'BookmarkId', u'Just a test website by Patrice')
        file.add(newBookmark)
        nodes = file.search('Patrice')
        assert len(nodes) == 2, 'Did not get correct number of search results'
        assert nodes[0] == newBookmark, 'Incorrect node returned by search'
        # append a folder to root
        newFolder = Folder(file.getFirst(), u'id', u'FolderTitle',
                           '', localtime(), False)
        file.add(newFolder)
        nodes = file.search('FolderTitle')
        assert len(nodes) == 1, 'Did not get correct number of search results'
        assert newFolder == nodes[0], 'Incorrect node returned by search'

    def testRemove(self):
        file = XBEL('bookmarks.xbel')
        nodes = file.search('pyunit')
        assert len(nodes) == 2, 'search did not return 2 results'
        assert file.remove(nodes[1]) == True, 'remove() did not succeed'
        nodes = file.search('pyunit')
        assert len(nodes) == 1, 'pyunit was not deleted'

        node = file.getFirst()
        file.remove(node)
        node = file.getFirst()
        assert file.remove(node) == False, 'remove() did succeed for root item'
        assert node != None, 'Could remove root'

        nodes = file.search('Python')
        assert len(nodes) == 3, 'search did not return 3 results'
        assert file.remove(nodes[0]) == True, 'remove() did not succeed'
        nodes = file.search('Python')
        assert len(nodes) == 0, 'Python folder was not deleted'

    def testMove(self):
        file = XBEL('bookmarks.xbel')
        nodes = file.search('Python Subfolder')
        file.move(nodes[0], file.getFirst())
        nodes = file.search('Python Subfolder')
        assert nodes[0].getParent().getTitle() == 'PyUnit test bookmarks'
        assert nodes[0].getParent() == file.getFirst()

    def testObserver(self):
        class FormatObserverImpl(FormatObserver):
            def __init__(self):
                self.__hasCreated_ = False
                self.__hasChanged_ = False
                self.__hasDeleted_ = False
                self.__hasMoved_ = False
                self.__node_ = None
            def created(self, node):
                self.__hasCreated_ = True
                self.__node_ = node
            def changed(self, node):
                self.__hasChanged_ = True
                self.__node_ = node
            def moved(self, node):
                self.__hasMoved_ = True
                self.__node_ = node
            def deleted(self, node):
                self.__hasDeleted_ = True
                self.__node_ = node
            def getCreated(self):
                return self.__hasCreated_
            def getChanged(self):
                return self.__hasChanged_
            def getMoved(self):
                return self.__hasMoved_
            def getDeleted(self):
                return self.__hasDeleted_
            def getNode(self):
                return self.__node_

        file = XBEL('bookmarks.xbel')
        observer = FormatObserverImpl()
        assert observer.getCreated() == False
        assert observer.getChanged() == False
        assert observer.getMoved() == False
        assert observer.getDeleted() == False
        file.subscribe(observer)
        nodes = file.search('python')
        newBookmark = Bookmark(nodes[3], u'http://www.patrice.ch/',
                               u'BookmarkId', u'Just a test website by Patrice')
        file.add(newBookmark)
        assert observer.getCreated()
        assert observer.getChanged() == False
        assert observer.getMoved() == False
        assert observer.getDeleted() == False

        file = XBEL('bookmarks.xbel')
        observer = FormatObserverImpl()
        file.subscribe(observer)
        nodes = file.search('Python Subfolder')
        nodes[0].setTitle('New title')
        assert observer.getCreated() == False
        assert observer.getChanged()
        assert observer.getMoved() == False
        assert observer.getDeleted() == False

        file = XBEL('bookmarks.xbel')
        observer = FormatObserverImpl()
        file.subscribe(observer)
        nodes = file.search('Python Subfolder')
        file.move(nodes[0], file.getFirst())
        assert observer.getCreated() == False
        assert observer.getChanged() == False
        assert observer.getMoved()
        assert observer.getDeleted() == False

        file = XBEL('bookmarks.xbel')
        observer = FormatObserverImpl()
        file.subscribe(observer)
        nodes = file.search('pyunit')
        assert file.remove(nodes[1]) == True, 'remove() did not succeed'
        assert observer.getCreated() == False
        assert observer.getChanged() == False
        assert observer.getMoved() == False
        assert observer.getDeleted()

    def testSave(self):
        file = XBEL()
        root = file.getFirst()
        root.setTitle(u'Test file')
        folder1 = Folder(root, u'1', u'New folder')
        bookmark1 = Bookmark(folder1, u'http://www.patrice.ch/', u'2', u'pt')
        bookmark2 = Bookmark(folder1, u'http://bkm.sf.net/', u'3', u'SF')
        bookmark3 = Bookmark(root, u'http://bible.gospelcom.net/', u'4', u'bible')
        file.add(folder1)
        file.add(bookmark1)
        file.add(bookmark2)
        file.add(bookmark3)
        file.save('tmp.xbel')

        file = XBEL('tmp.xbel')
        root = file.getFirst()
        assert root.getTitle() == u'Test file'
        folder1 = root.getFirstChild()
        assert folder1 != None
        assert folder1.getId() == u'1'
        assert folder1.getTitle() == u'New folder'
        bookmark1 = folder1.getFirstChild()
        assert bookmark1 != None
        assert bookmark1.getHref() == u'http://www.patrice.ch/'
        assert bookmark1.getId() == u'2'
        assert bookmark1.getTitle() == u'pt'
        bookmark2 = bookmark1.getNextSibling()
        assert bookmark2 != None
        assert bookmark2.getHref() == u'http://bkm.sf.net/'
        assert bookmark2.getId() == u'3'
        assert bookmark2.getTitle() == u'SF'
        bookmark3 = folder1.getNextSibling()
        assert bookmark3 != None
        assert bookmark3.getHref() == u'http://bible.gospelcom.net/'
        assert bookmark3.getId() == u'4'
        assert bookmark3.getTitle() == u'bible'
        os.unlink('tmp.xbel')

    def testOpenSave(self):
        file = XBEL('bookmarks.xbel')
        file.save('tmp.xbel')
        file = XBEL('tmp.xbel')
        nodes = file.search('python')
        assert len(nodes) == 4, 'Did not get correct number of search results'
        assert nodes[0].getTitle() == 'Python', 'Invalid search result returned'
        assert nodes[1].getTitle() == 'python.org', 'Invalid search result returned'
        assert nodes[1].getHref() == 'http://www.python.org', 'Invalid search result returned'
        assert nodes[2].getTitle() == 'pyunit', 'Invalid search result returned'
        assert nodes[3].getTitle() == 'Python Subfolder', 'Invalid search result returned'

        nodes = file.search('unit')
        assert len(nodes) == 2, 'Did not get correct number of search results'
        assert nodes[0].getTitle() == 'PyUnit test bookmarks'
        assert nodes[1].getTitle() == 'pyunit'
        assert nodes[1].getHref() == 'http://pyunit.sourceforge.net'
        assert nodes[1].getDesc() == 'Python unit testing.'
        assert len(nodes[1].getMetadata()) == 2
        assert nodes[1].getMetadata()[0].getOwner() == \
               'http://galeon.sourceforge.net/'
        assert nodes[1].getMetadata()[0].getValue('time_visited') == \
               '1012330509'
        assert nodes[1].getMetadata()[1].getOwner() == \
               'http://bkm.sf.net/'
        assert nodes[1].getMetadata()[1].getValue('keywords') == \
               'unit testing, junit, xunit, extreme programming'
        assert nodes[1].getMetadata()[1].getValue('http_status') == \
               '200'

        nodes = file.search('gutelaune')
        assert len(nodes) == 0, 'Returned result for inexistient item'
        os.unlink('tmp.xbel')

    def testParse1(self):
        """Use the XBel.parse() method to parse a tree"""
        data = '<xbel version="1.0" id="id2" added="a">\n'
        data += '<title>t</title>\n'
        data += '<desc>d</desc>\n'
        data += '<info>\n'
        data += '<metadata owner="http://bkm.sf.net/">\n'
        data += '<rating>3</rating>\n'
        data += '</metadata>\n'
        data += '</info>\n'
        data += '<folder id="id3" folded="yes">\n'
        data += '<title>subfolder</title>\n'
        data += '</folder>\n'
        data += '</xbel>\n'

        xbel = XBEL()
        xbel.parse(data)

        root = xbel.getFirst()
        assert root.getTitle() == u"t"
        assert root.getDesc() == u"d"
        assert root.getFirstChild().getTitle() == u"subfolder"
        assert root.getFirstChild().getFolded() == True
        assert root.getFirstChild().getId() == u"id3"

    def testParse2(self):
        """Use the XBel.parse() method to parse a tree. This is two
        folders on the same level with no root folder. A root element
        should be automatically created."""

        data = '<folder id="id2">\n'
        data += '<title>t</title>\n'
        data += '<desc>d</desc>\n'
        data += '</folder>\n'
        data += '<folder id="id3" folded="yes">\n'
        data += '<title>subfolder</title>\n'
        data += '</folder>\n'

        xbel = XBEL()
        xbel.parse(data)

        root = xbel.getFirst()
        assert root.getTitle() == u"Root"
        f = root.getFirstChild()
        assert f.getTitle() == u"t"
        assert f.getId() == u"id2"
        assert f.getDesc() == u"d"
        f = f.getNextSibling()
        assert f.getTitle() == u"subfolder"
        assert f.getFolded() == True
        assert f.getId() == u"id3"

    def testParse3(self):
        """Use the XBel.parse() method to parse a tree. This is one
        bookmark with no root folder"""

        data = '<bookmark id="bkm" href="http://bkm.sourceforge.net/">\n'
        data += '<title>bkm</title>\n'
        data += '<desc>Website for vbBKM and pyBKM</desc>\n'
        data += '</bookmark>\n'

        xbel = XBEL()
        xbel.parse(data)

        root = xbel.getFirst()
        assert root.getTitle() == u"Root"
        b = root.getFirstChild()
        assert b.getTitle() == u"bkm"
        assert b.getId() == u"bkm"
        assert b.getDesc() == u"Website for vbBKM and pyBKM"
        assert b.getHref() == u"http://bkm.sourceforge.net/"
