# Author:    Patrice Neff
# Purpose:   Unittests for the Data.Node package
#
# History:
#   - 2003-05-07  created, pn

import unittest
import os
from Node import *

class NodeTest(unittest.TestCase):
    def testDefaultConstructor(self):
        node = DataNode(None)
        assert node.getParent() == None
        assert node.getId() == ''
        assert node.getTitle() == ''
        assert node.getDesc() == ''
        assert node.getAdded() == None

    def testConstructor(self):
        parent = Folder(None, 'id', 'title', 'desc', 'time')
        node = DataNode(parent, 'nodeId', 'nodeTitle', 'nodeDesc', 'nodeTime')
        assert node.getParent() == parent
        assert node.getId() == 'nodeId'
        assert node.getTitle() == 'nodeTitle'
        assert node.getDesc() == 'nodeDesc'
        assert node.getAdded() == None

    def testObserver(self):
        class NodeObserverImpl(NodeObserver):
            def __init__(self):
                self.__hasChanged_ = False
                self.__node_ = None
            def changed(self, node):
                self.__hasChanged_ = True
                self.__node_ = node
            def getChanged(self):
                return self.__hasChanged_
            def getNode(self):
                return self.__node_

        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        assert observer.getChanged() == False
        node.setParent(Folder(None, 'newnode'))
        assert observer.getChanged() == True
        assert observer.getNode() == node
        
        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        node.setId('newid')
        assert observer.getChanged() == True

        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        node.setTitle('newtitle')
        assert observer.getChanged() == True

        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        node.setDesc('newdescription')
        assert observer.getChanged() == True

        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        node.setRating(5)
        assert observer.getChanged() == True

        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        node.setAdded('newtime')
        assert observer.getChanged() == True

        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        mdata = Metadata()
        mdata.setValue('key', 'hi')
        node.addMetadata(mdata)
        assert observer.getChanged() == True

        node = DataNode(None)
        observer = NodeObserverImpl()
        node.subscribe(observer)
        node.removeMetadata(mdata)
        assert observer.getChanged() == True

    def testMetadata(self):
        node = DataNode(None)
        mdata = Metadata()
        mdata.setValue('key', 'hi')
        mdata.setValue('duda', 'sodale')
        node.addMetadata(mdata)
        mdata = Metadata('http//www.patrice.ch/')
        mdata.setValue('host', 'www.patrice.ch')
        mdata.setValue('host', 'patrice.ch')
        node.addMetadata(mdata)

        mdata = node.getMetadata()
        assert len(mdata) == 2
        assert mdata[0].getValue('key') == 'hi'
        assert mdata[0].getValue('duda') == 'sodale'
        assert mdata[1].getValue('host') == 'patrice.ch'

class FolderTest(unittest.TestCase):
    def testInheritance(self):
        folder = Folder(None, 'id', 'title', 'description', '2004-04-09T21:39:00+00:00')
        assert folder.getId() == 'id', 'Constructor does not set id'
        assert folder.getTitle() == 'title', 'Constructor does not set title'
        assert folder.getDesc() == 'description', 'Constructor does not set description'
        a = folder.getAdded()
        assert a[0] == 2004 and a[1] == 4, 'Constructor does not set added'

    def testEquality(self):
        import time
        t = time.localtime(1082340410)
        folder0 = Folder(None, 'id1', 'My title', 'New description', t, False)
        folder1 = Folder(None, 'id1', 'My title', 'New description', t, False)
        assert folder0 == folder1, '== for equal objects fails'

        folder1 = Folder(Folder(None), 'id2', 'My title', 'New description', t, False)
        assert folder0 != folder1, 'Returned == on different parents'
        folder1 = Folder(None, 'id2', 'My title', 'New description', t, False)
        assert folder0 != folder1, 'Returned == on different ID'
        folder1 = Folder(None, 'id1', 'Other title', 'New description', t, False)
        assert folder0 != folder1, 'Returned == on different title'
        folder1 = Folder(None, 'id1', 'My title', 'Other description', t, False)
        assert folder0 != folder1, 'Returned == on different description'
        folder1 = Folder(None, 'id1', 'My title', 'New description', '2004-04-09T21:39:00+00:00', False)
        assert folder0 != folder1, 'Returned == on different time'
        folder1 = Folder(None, 'id1', 'My title', 'New description', t, True)
        assert folder0 != folder1, 'Returned == on different Folded attribute'

    def testFolded(self):
        folder = Folder(None)
        assert folder.getFolded(), 'Default for folded must be true'
        
        folder = Folder(None, 'id', 'title', 'description', 'added', False)
        assert folder.getFolded() == False, 'Folded was not set correctly in constructor'
        
        folder.setFolded(True)
        assert folder.getFolded(), 'setFolded did not set folded value'
        folder.setFolded(False)
        assert folder.getFolded() == False, 'setFolded did not set folded value'

    def testRating(self):
        folder = Folder(None)
        assert folder.getRating() == 0
        folder.setRating(3)
        assert folder.getRating() == 3
        folder.setRating("a")
        assert folder.getRating() == 0
        folder.setRating(-50)
        assert folder.getRating() == 0

    def testCanContainNodes(self):
        folder = Folder(None)
        assert folder.canContainNodes(), 'Folder must be able to contain nodes'

    def testCanExecute(self):
        folder = Folder(None)
        assert folder.canExecute() == False, 'Bookmark must be executable'

    def testIsDescendantOf(self):
        folder0 = Folder(None, 'id1', 'My title', 'New description')
        folder1 = Folder(folder0, 'id1', 'My title', 'New description')
        bk0 = Bookmark(folder0, 'http://bkm.sf.net/', 'id', 'title', 'description')
        bk1 = Bookmark(folder1, 'http://bkm.sf.net/', 'id', 'title', 'description')

        assert folder1.containsDescendant(folder0) == False
        assert folder0.containsDescendant(folder1)
        assert folder1.containsDescendant(bk1)
        assert folder1.containsDescendant(bk0) == False
        assert folder0.containsDescendant(bk1)
        assert folder0.containsDescendant(bk0)

    def testWriteXbel1(self):
        """Write one folder to XBEL and verify output."""
        # construct tree
        file = os.tmpfile()
        folder = Folder(None, u'id1', u'title', u'desc', u'1082382793', True)

        # write
        folder.writeXbel(file)

        # verify output
        output = ''
        file.seek(0)
        while True:
            buf = file.read()
            if buf == '':
                break
            output += buf
        outputWished = '<xbel version="1.0" id="id1" added="1082382793" folded="yes">\n'
        outputWished += '<title>title</title>\n'
        outputWished += '<desc>desc</desc>\n'
        outputWished += '</xbel>\n'
        assert output == outputWished
        assert output == folder.getXbel()

    def testWriteXbel2(self):
        """Write folder with subfolder to XBEL and verify output."""
        # construct tree
        file = os.tmpfile()        
        folder = Folder(None, u'id2', u't', u'd', u'', False)
        folder.setRating(3)
        folder1 = Folder(folder, u'id3', u'subfolder')
        folder.setFirstChild(folder1)

        # write
        folder.writeXbel(file)

        # verify output
        output = ''
        file.seek(0)
        while True:
            buf = file.read()
            if buf == '':
                break
            output += buf
        outputWished = '<xbel version="1.0" id="id2">\n'
        outputWished += '<title>t</title>\n'
        outputWished += '<desc>d</desc>\n'
        outputWished += '<info>\n'
        outputWished += '<metadata owner="http://bkm.sf.net/">\n'
        outputWished += '<rating>3</rating>\n'
        outputWished += '</metadata>\n'
        outputWished += '</info>\n'
        outputWished += '<folder id="id3" folded="yes">\n'
        outputWished += '<title>subfolder</title>\n'
        outputWished += '</folder>\n'
        outputWished += '</xbel>\n'
        assert output == outputWished
        assert output == folder.getXbel()

    def testWriteXbel3(self):
        """Write folder with subfolders and bookmarks to XBEL and verify."""
        # construct tree
        import time
        file = os.tmpfile()
        folder = Folder(None, u'id2', u't', u'd', time.localtime(1082340410), False)
        folder1 = Folder(folder, u'id3', u'subfolder')
        folder.setFirstChild(folder1)
        bk1 = Bookmark(folder, u'url')
        folder1.setNextSibling(bk1)
        bk2 = Bookmark(folder, u'href')
        folder1.setFirstChild(bk2)
        sep = Separator(folder)
        bk1.setNextSibling(sep)
        folder2 = Folder(folder)
        sep.setNextSibling(folder2)
        folder2.setFolded(False)
        # folder
        #  - folder1
        #    - bk2
        #  - bk1
        #  - separator
        #  - folder2

        # write
        folder.writeXbel(file)

        # verify output
        output = ''
        file.seek(0)
        while True:
            buf = file.read()
            if buf == '':
                break
            output += buf
        outputWished = '<xbel version="1.0" id="id2" added="1082340410">\n'
        outputWished += '<title>t</title>\n'
        outputWished += '<desc>d</desc>\n'
        outputWished += '<folder id="id3" folded="yes">\n'
        outputWished += '<title>subfolder</title>\n'
        outputWished += '<bookmark href="href">\n</bookmark>\n'
        outputWished += '</folder>\n'
        outputWished += '<bookmark href="url">\n</bookmark>\n'
        outputWished += '<separator/>\n'
        outputWished += '<folder>\n</folder>\n'
        outputWished += '</xbel>\n'
        assert output == outputWished
        assert output == folder.getXbel()

class BookmarkTest(unittest.TestCase):
    def testInheritance(self):
        """Test if DataNode methods are inherited"""
        import time
        t = time.localtime(1082340410)
        bookmark = Bookmark(None, 'href', 'id', 'title', 'description', t)
        assert bookmark.getId() == 'id', 'Constructor does not set id'
        assert bookmark.getTitle() == 'title', 'Constructor does not set title'
        assert bookmark.getDesc() == 'description', 'Constructor does not set description'
        assert bookmark.getAdded() == t, 'Constructor does not set added'

    def testHref(self):
        try:
            bookmark = Bookmark(None)
        except:
            pass
        else:
            self.fail('Bookmark accepts empty parameter list')
        
        bookmark = Bookmark(None, 'href', 'id')
        assert bookmark.getHref() == 'href', 'Constructor did not set href'

        bookmark = Bookmark(None, 'http://www.patrice.ch/', 'id', 'Title', 'Description', "09.02.1983 09:10:15")
        assert bookmark.getHref() == 'http://www.patrice.ch/', 'Constructor did not set href'

        bookmark.setHref('bkm.sf.net')
        assert bookmark.getHref() == 'bkm.sf.net', 'setHref did not set href'

    def testRating(self):
        bookmark = Bookmark(None, 'href')
        assert bookmark.getRating() == 0
        bookmark.setRating(3)
        assert bookmark.getRating() == 3
        bookmark.setRating("a")
        assert bookmark.getRating() == 0
        bookmark.setRating(-1)
        assert bookmark.getRating() == 0

    def testModified(self):
        import time
        
        bookmark = Bookmark(None, 'href')
        assert bookmark.getModified() == None, 'Invalid default value for modified'
        t = time.localtime(1082340410)
        bookmark.setModified(t)
        assert bookmark.getModified() == t, 'setModified did not set value'

    def testVisited(self):
        import time
        
        bookmark = Bookmark(None, 'href')
        assert bookmark.getVisited() == None, 'Invalid default value for visited'
        t = time.localtime(1082340410)
        bookmark.setVisited(t)
        assert bookmark.getVisited() == t, 'setVisited did not set value'

    def testEquality(self):
        import time
        
        t = time.localtime(1082340410)
        bk0 = Bookmark(None, 'http://bkm.sf.net/', 'id', 'title', 'description', t)
        bk1 = Bookmark(None, 'http://bkm.sf.net/', 'id', 'title', 'description', t)
        assert bk0 == bk1, '== for equal objects fails'

        bk1 = Bookmark(None, 'http://bkm.sf.net/', 'id', 'title', 'description', t)
        assert bk0 == bk1, '== for equal objects fails'

    def testCanContainNodes(self):
        bookmark = Bookmark(None, 'href')
        assert bookmark.canContainNodes() == False, 'Bookmark must not be able to contain nodes'

    def testCanExecute(self):
        bookmark = Bookmark(None, 'href')
        assert bookmark.canExecute() == True, 'Bookmark must be executable'

    def testIsDescendantOf(self):
        folder0 = Folder(None, 'id1', 'My title', 'New description')
        folder1 = Folder(folder0, 'id1', 'My title', 'New description')
        bk0 = Bookmark(folder0, 'http://bkm.sf.net/', 'id', 'title', 'description')
        bk1 = Bookmark(folder1, 'http://bkm.sf.net/', 'id', 'title', 'description')

        assert bk1.containsDescendant(folder0) == False, 'bookmark must not contain other nodes'
        assert bk0.containsDescendant(folder1) == False, 'bookmark must not contain other nodes'
        assert bk1.containsDescendant(folder1) == False, 'bookmark must not contain other nodes'
        assert bk1.containsDescendant(bk0) == False, 'bookmark must not contain other nodes'
        assert bk0.containsDescendant(bk1) == False, 'bookmark must not contain other nodes'

    def testWriteXbel1(self):
        """Write bookmark to XBEL and verify output."""
        import time
        file = os.tmpfile()        
        b = Bookmark(None, u'href', u'id', u'title', u'desc', time.localtime(1084340410))
        b.setRating(2)

        # write
        b.writeXbel(file)

        # verify output
        output = ''
        file.seek(0)
        while True:
            buf = file.read()
            if buf == '':
                break
            output += buf
        outputWished = '<bookmark href="href" id="id" added="1084340410">\n'
        outputWished += '<title>title</title>\n'
        outputWished += '<desc>desc</desc>\n'
        outputWished += '<info>\n'
        outputWished += '<metadata owner="http://bkm.sf.net/">\n'
        outputWished += '<rating>2</rating>\n'
        outputWished += '</metadata>\n'
        outputWished += '</info>\n'
        outputWished += '</bookmark>\n'
        assert output == outputWished
        assert output == b.getXbel()

    def testWriteXbel2(self):
        """Write bookmark with metadata to XBEL and verify output."""
        file = os.tmpfile()        
        b = Bookmark(None, u'href', u'id', u'title', u'desc')
        b.setRating(2)
        m = Metadata(u'http://slashdot.org/')
        m.setValue(u'user', u'username')
        m.setValue(u'pass', u'myPass')
        m.setValue(u'strange', u'<>sab')
        m.setValue(u'Feed', u'http://slashdot.org/slashdot.rdf')
        b.addMetadata(m)

        # write
        b.writeXbel(file)

        # verify output
        output = ''
        file.seek(0)
        while True:
            buf = file.read()
            if buf == '':
                break
            output += buf
        outputWished = '<bookmark href="href" id="id">\n'
        outputWished += '<title>title</title>\n'
        outputWished += '<desc>desc</desc>\n'
        outputWished += '<info>\n'
        outputWished += '<metadata owner="http://bkm.sf.net/">\n'
        outputWished += '<rating>2</rating>\n'
        outputWished += '</metadata>\n'
        outputWished += '<metadata owner="http://slashdot.org/">\n'
        outputWished += '<Feed>http://slashdot.org/slashdot.rdf</Feed>\n'
        outputWished += '<strange>&lt;&gt;sab</strange>\n'
        outputWished += '<user>username</user>\n'
        outputWished += '<pass>myPass</pass>\n'
        outputWished += '</metadata>\n'
        outputWished += '</info>\n'
        outputWished += '</bookmark>\n'
        assert output == outputWished
        assert output == b.getXbel()


class SeparatorTest(unittest.TestCase):
    def testInheritance(self):
        """Test if Node methods are inherited"""
        s = Separator(None)
        f = Folder(None, '', 'test')
        assert s.getParent() == None
        s.setParent(f)
        assert s.getParent().getTitle() == 'test'
        s = Separator(f)
        assert s.getParent().getTitle() == 'test'

    def testEquality(self):
        s0 = Separator(None)
        s1 = Separator(None)
        assert s0 == s1

        f = Folder(None)
        s0.setParent(f)
        assert s0 != s1

        s1.setParent(f)
        assert s0 == s1

    def testCanContainNodes(self):
        s = Separator(None)
        assert s.canContainNodes() == False, 'Separator must not be able to contain nodes'

    def testCanExecute(self):
        s = Separator(None)
        assert s.canExecute() == False, 'Separator must not be executable'

    def testIsDescendantOf(self):
        folder0 = Folder(None, 'id1', 'My title', 'New description')
        folder1 = Folder(folder0, 'id1', 'My title', 'New description')
        bk0 = Bookmark(folder0, 'http://bkm.sf.net/', 'id', 'title', 'description')
        bk1 = Bookmark(folder1, 'http://bkm.sf.net/', 'id', 'title', 'description')
        s0 = Separator(folder0)
        s1 = Separator(folder1)

        assert s1.containsDescendant(folder0) == False, 'Separator must not contain other nodes'
        assert s0.containsDescendant(folder1) == False, 'Separator must not contain other nodes'
        assert s1.containsDescendant(folder1) == False, 'Separator must not contain other nodes'
        assert s1.containsDescendant(bk0) == False, 'Separator must not contain other nodes'
        assert s0.containsDescendant(bk1) == False, 'Separator must not contain other nodes'
        assert s1.containsDescendant(s0) == False, 'Separator must not contain other nodes'
        assert s0.containsDescendant(s1) == False, 'Separator must not contain other nodes'

    def testWriteXbel(self):
        """Write XBEL to a temporary file and test if the output is
        correct."""
        f = os.tmpfile()        
        sep = Separator(None)
        sep.writeXbel(f)

        output = ''
        f.seek(0)
        while True:
            buf = f.read()
            if buf == '':
                break
            output += buf
        assert output == "<separator/>\n"
        assert output == sep.getXbel()
