"""
@author:  Patrice Neff
@since:   2004-02-14
@license: GNU General Public License
@version: $Revision: 1.25 $

Display the main window. This is the window loaded on program startup.
"""

import os.path
from wxPython.wx import *
from Data.Formats import XBEL
from Data.Node import NodeObserver, Bookmark
from Gui.Tree import Tree, TreeObserver, MainTreeHandler
from Config.RecentFiles import RecentFiles
from Gui.Helper import QuestionDialog, uconv

ID_MNU_NEW = 101
ID_MNU_OPEN = 102
ID_MNU_SAVE = 103
ID_MNU_SAVEAS = 104
ID_MNU_CLOSE = 105
ID_MNU_EXIT  = 106
ID_MNU_RECENT1 = 107
ID_MNU_RECENT2 = 108
ID_MNU_RECENT3 = 109
ID_MNU_RECENT4 = 110
ID_MNU_NEW_BOOKMARK = 201
ID_MNU_NEW_FOLDER = 202
ID_MNU_NEW_SEPARATOR = 203
ID_MNU_DELETE = 204
ID_MNU_COPY = 205
ID_MNU_CUT = 206
ID_MNU_PASTE = 207
ID_BUT_SAVE = 501

class MainFrame(wxFrame):
    """The main window of pyBKM.

    @since: 2004-02-14
    """
    def __init__(self, title):
        """Initialize the main pyBKM window.

        @param title: Window title
        @type title: String
        """
        # Initialize local variables
        self.__filename = None
        self.__directory = ""
        self.__recent_separator = False # Set to True if the separator for recent files
                                        # has been created

        # Configuration
        self.__config = wxConfig(style=wxCONFIG_USE_LOCAL_FILE)
        self.__recentfiles = RecentFiles(self.__config)

        # Setup up window
        wxFrame.__init__(self, None, -1, title,
                         wxDefaultPosition, wxSize(400, 420))

        # Create splitter window
        splitter = wxSplitterWindow (self, -1, style=wxNO_3D|wxSP_3D)
        splitter.SetMinimumPaneSize(200)
        self.__tree = Tree(splitter, wxNewId(), None)
        self.__editor = MainWindowEditFrame(splitter, -1)
        self.__tree.subscribe(self.__editor)
        self.__tree.subscribe(MainTreeHandler())
        splitter.SplitVertically (self.__tree, self.__editor, -300)

        # Set up the menu bar
        self.CreateMenuBar()

        # locale
        wxFont_SetDefaultEncoding(wxFONTENCODING_ISO8859_15)

        if len(self.__recentfiles.getList()) > 0:
            for file in self.__recentfiles.getList():
                if os.path.exists(file):
                    self.OpenFile(file)
                    break
        else:
            self.OpenFile('')
        self.Show(true)

    #
    # Menu
    #
    def CreateMenuBar(self):
        """Create a menu bar and associate the menu items with the
        event handler functions."""
        mnu_file = wxMenu()
        self.__mnu_file = mnu_file
        mnu_file.Append(ID_MNU_NEW, _("&New"), _("New database"))
        mnu_file.Append(ID_MNU_OPEN, _("&Open\tCtrl+O"), _("Open database"))
        mnu_file.Append(ID_MNU_SAVE, _("&Save\tCtrl+S"), _("Save database"))
        mnu_file.Append(ID_MNU_SAVEAS, _("Save &as\tCtrl+Shift+S"), _("Save database with new name"))
        mnu_file.Append(ID_MNU_CLOSE, _("&Close\tCtrl+W"), _("Close database"))
        mnu_file.AppendSeparator()
        mnu_file.Append(ID_MNU_EXIT, _("E&xit\tCtrl+Q"), _("Exit pyBKM"))
        self.CreateRecentFileList()

        mnu_bookmark = wxMenu()
        self.__mnu_bookmark = mnu_bookmark
        mnu_bookmark.Append(ID_MNU_NEW_BOOKMARK, _("New &Bookmark\tCtrl+N"), _("Create new bookmark"))
        mnu_bookmark.Append(ID_MNU_NEW_FOLDER, _("New &Folder\tCtrl+F"), _("Create new folder"))
        mnu_bookmark.Append(ID_MNU_NEW_SEPARATOR, _("New &Separator\tCtrl+P"), _("Create new separator"))
        mnu_bookmark.Append(ID_MNU_DELETE, _("Delete\tCtrl+D"), _("Delete selected node"))
        mnu_bookmark.AppendSeparator()
        mnu_bookmark.Append(ID_MNU_COPY, _("Copy\tCtrl+C"), _("Copy selected node to clipboard"))
        mnu_bookmark.Append(ID_MNU_CUT, _("Cut\tCtrl+X"), _("Cut selected node to clipboard"))
        mnu_bookmark.Append(ID_MNU_PASTE, _("Paste\tCtrl+V"), _("Paste node from clipboard"))
        
        menubar = wxMenuBar()
        menubar.Append(mnu_bookmark, _("&Bookmark"))
        menubar.Append(mnu_file, _("&File"))
        self.SetMenuBar(menubar)
        self.ActivateMenu()
        
        EVT_MENU(self, ID_MNU_NEW, self.OnMenuNew)
        EVT_MENU(self, ID_MNU_OPEN, self.OnMenuOpen)
        EVT_MENU(self, ID_MNU_SAVE, self.OnMenuSave)
        EVT_MENU(self, ID_MNU_SAVEAS, self.OnMenuSaveAs)
        EVT_MENU(self, ID_MNU_CLOSE, self.OnMenuClose)
        EVT_MENU(self, ID_MNU_EXIT, self.OnMenuExit)
        EVT_MENU(self, ID_MNU_NEW_BOOKMARK, self.OnMenuNewBookmark)
        EVT_MENU(self, ID_MNU_NEW_FOLDER, self.OnMenuNewFolder)
        EVT_MENU(self, ID_MNU_NEW_SEPARATOR, self.OnMenuNewSeparator)
        EVT_MENU(self, ID_MNU_DELETE, self.OnMenuDelete)
        EVT_MENU(self, ID_MNU_COPY, self.OnMenuCopy)
        EVT_MENU(self, ID_MNU_CUT, self.OnMenuCut)
        EVT_MENU(self, ID_MNU_PASTE, self.OnMenuPaste)

    def CreateRecentFileList(self):
        """Create the menu items for the recent files list in the file
        menu.
        """
        mnu = self.__mnu_file

        # remove previous entries
        for id in range(ID_MNU_RECENT1, ID_MNU_RECENT4+1):
            if mnu.FindItemById(id) is not None:
                mnu.Delete(id)

        # write new list
        for i, item in enumerate(self.__recentfiles.getList()):
            if item != "":
                if not self.__recent_separator:
                    mnu.InsertSeparator(mnu.GetMenuItemCount()-2)
                    self.__recent_separator = True
                mnu.Insert(mnu.GetMenuItemCount()-2, i+ID_MNU_RECENT1, "&" +
                           str(i+1) + ": " + os.path.basename(item))

        # event handlers
        for id in range(ID_MNU_RECENT1, ID_MNU_RECENT4+1):
            EVT_MENU(self, id, self.OnMenuRecent)

    def ActivateMenu(self):
        """Enable and disable menu items depending on the current
        application state. Must be called whenever any action is
        executed that could change the state of some menu items."""        
        fileOpen = self.__filename is not None
        nodeSelected = self.__tree.getSelectedNode() != None

        m = self.GetMenuBar()
        m.Enable(ID_MNU_SAVE, fileOpen)
        m.Enable(ID_MNU_CLOSE, fileOpen)
        m.Enable(ID_MNU_NEW_BOOKMARK, fileOpen and nodeSelected)
        m.Enable(ID_MNU_NEW_FOLDER, fileOpen and nodeSelected)
        m.Enable(ID_MNU_NEW_SEPARATOR, fileOpen and nodeSelected)
        m.Enable(ID_MNU_DELETE, fileOpen and nodeSelected)

    def OpenFile(self, filename):
        """Open a file.

        @param filename: File to open. Create new file if filename is
                         empty.
        @type filename: String
        """
        self.__filename = filename
        self.__directory = os.path.dirname(filename)
        self.__tree.setDatabase(XBEL(self.__filename))
        self.__recentfiles.addFile(filename)
        self.CreateRecentFileList()
        self.ActivateMenu()

    #
    # menu handlers
    #
    def OnMenuNew(self, event):
        """Handle menu command New."""
        self.OpenFile('')

    def OnMenuOpen(self, event):
        """Handle menu command Open."""
        dlg = wxFileDialog(self, _("Choose a file"), self.__directory, "", "*.xbel", wxOPEN)
        if dlg.ShowModal() == wxID_OK:
            self.OpenFile(dlg.GetPath())
        dlg.Destroy()        
        self.ActivateMenu()

    def OnMenuSave(self, event):
        """Handle menu command Save."""
        if self.__filename is None or self.__filename == '':
            self.OnMenuSaveAs(event)
        else:
            self.__tree.getDatabase().save(self.__filename)
        self.ActivateMenu()

    def OnMenuSaveAs(self, event):
        """Handle menu command Save As."""
        dlg = wxFileDialog(self, _("Save as"), self.__directory, "", "*.xbel", wxSAVE)
        if dlg.ShowModal() == wxID_OK:
            self.__directory = dlg.GetDirectory()
            self.__filename = dlg.GetPath()
            self.OnMenuSave(event)
        dlg.Destroy()
        self.ActivateMenu()

    def OnMenuClose(self, event):
        """Handle menu command Close."""
        if self.__tree.isDirty():
            q = QuestionDialog(_("This bookmark database has been changed. Do you want to save your changes?"), _("Save changes?"))
            if q.getAnswer() == wxID_YES:
                self.OnMenuSave(event)
        
        self.OnMenuNew(event)
        self.ActivateMenu()

    def OnMenuExit(self, event):
        """Handle menu command Exit."""
        if self.__tree.isDirty():
            q = QuestionDialog(_("This bookmark database has been changed. Do you want to save your changes?"), _("Save changes?"))
            if q.getAnswer() == wxID_YES:
                self.OnMenuSave(event)
        
        self.Close(True)

    def OnMenuRecent(self, event):
        """Handle menu command Recently opened files."""
        files = self.__recentfiles.getList()
        file = files[event.GetId()-ID_MNU_RECENT1]
        if file != None and file != '':
            self.OpenFile(file)

    def OnMenuNewBookmark(self, event):
        """Handle menu command New Bookmark."""
        self.__tree.NewBookmark()

    def OnMenuNewFolder(self, event):
        """Handle menu command New Folder."""
        self.__tree.NewFolder()

    def OnMenuNewSeparator(self, event):
        """Handle menu command New Separator."""
        self.__tree.NewSeparator()

    def OnMenuDelete(self, event):
        """Handle menu command Delete."""
        self.__tree.DeleteSelected()

    def OnMenuCopy(self, event):
        """Handle menu command Copy."""
        self.__tree.CopyNode()

    def OnMenuCut(self, event):
        """Handle menu command Cut."""
        self.__tree.CutNode()

    def OnMenuPaste(self, event):
        """Handle menu command Paste."""
        self.__tree.PasteNode()

class MainWindowEditFrame(wxPanel,TreeObserver,NodeObserver):
    """
    Edit box to edit attributes of the current node.

    @since: 2004-02-14
    """
    def __init__(self, parent, ID):
        """Constructor. Initialize and paint the edit frame.

        @param parent: Window to paint the tree on
        @type parent: L{wxWindow}
        @param ID: ID of this window
        @type ID: L{wxWindowID}
        """
        self.__node = None
        
        wxPanel.__init__(self, parent, ID)

        self.__box = wxStaticBox(self, -1, _("Edit"))
        sizer = wxStaticBoxSizer(self.__box, wxVERTICAL)

        self.__txttitle = wxTextCtrl(self, -1)
        sizer.Add(wxStaticText(self, -1, _("Title")), 0, wxEXPAND|wxTOP, 10)
        sizer.Add(self.__txttitle, 0, wxEXPAND)

        self.__lblurl = wxStaticText(self, -1, _("URL"))
        self.__txturl = wxTextCtrl(self, -1)
        sizer.Add(self.__lblurl, 0, wxEXPAND|wxTOP, 10)
        sizer.Add(self.__txturl, 0, wxEXPAND)

        self.__chorating = wxChoice(self, -1, choices = [
            _("(not rated)"), _("Poor"), _("Average"), _("Good"), _("Great")
            ])
        sizer.Add(wxStaticText(self, -1, _("Rating")), 0, wxEXPAND|wxTOP, 10)
        sizer.Add(self.__chorating, 0, wxEXPAND)

        self.__txtdesc = wxTextCtrl(self, -1, style = wxTE_MULTILINE)
        sizer.Add(wxStaticText(self, -1, _("Description")), 0, wxEXPAND|wxTOP, 10)
        sizer.Add(self.__txtdesc, 1, wxEXPAND)

        sizer.Add(wxButton(self, ID_BUT_SAVE, _("&Save")), 0, wxALIGN_RIGHT|wxTOP, 10)
        EVT_BUTTON(self, ID_BUT_SAVE, self.OnButtonSave)

        self.SetSizer(sizer)

    def OnButtonSave(self, event):
        """Save the input of the text controls to the currently active
        node.
        """
        if self.__node is not None:
            self.__node.unsubscribe(self) # no change events while saving
            node = self.__node
            node.setTitle(uconv(self.__txttitle.GetValue()))
            node.setRating(self.__chorating.GetSelection())
            try:
                node.setHref(uconv(self.__txturl.GetValue()))
            except:
                pass # setHref() only works with bookmarks
            node.setDesc(uconv(self.__txtdesc.GetValue()))
            self.__node.subscribe(self)

    def changed(self, node):
        """The currently displayed node has been changed. This works,
        because we only subscribe to the current node. Redisplay the
        node's properties in the edit window. Function implemented for
        the L{NodeObserver} interface.

        @param node: Changed node
        @type node: L{Node}
        """
        self.selected(node)

    def selected(self, node):
        """A node has been selected. Display the selected node in the
        edit window and allow editing. Function implemented for the
        L{TreeObserver} interface.

        @param node: Selected node
        @type node: L{Node}
        """
        if self.__node is not None:
            self.__node.unsubscribe(self)
        try:
            title = node.getTitle()
        except:
            # separator
            node = None

        self.__node = node

        if node is None:
            self.__txttitle.SetValue('')
            self.__txttitle.Enable(False)
            self.__txturl.SetValue('')
            self.__txturl.Enable(False)
            self.__chorating.SetSelection(0)
            self.__chorating.Enable(False)
            self.__txtdesc.SetValue('')
            self.__txtdesc.Enable(False)
            self.__box.Enable(False)
        else:
            node.subscribe(self)
            self.__txttitle.SetValue(node.getTitle().encode("iso-8859-15"))
            if node.getRating() <= 4:
                self.__chorating.SetSelection(node.getRating())
            else:
                self.__chorating.SetSelection(0)
            self.__txtdesc.SetValue(node.getDesc().encode("iso-8859-15"))
            try:
                self.__txturl.SetValue(node.getHref().encode("iso-8859-15"))
            except:
                # getHref() only available for bookmarks
                self.__txturl.SetValue('')
            
            self.__lblurl.Enable(isinstance(node, Bookmark))
            self.__txturl.Enable(isinstance(node, Bookmark))
            self.__txttitle.Enable(True)
            self.__chorating.Enable(True)
            self.__txtdesc.Enable(True)
            self.__box.Enable(False)
            self.__box.Enable(True)
        

    def activated(self, bookmark):
        """A bookmark has been activated, i.e. double-clicked.
        Function implemented for the L{TreeObserver} interface.

        @param bookmark: Activated bookmark
        @type bookmark: L{Bookmark}
        """
        pass # event ignored
