"""
@author:  Patrice Neff
@since:   2004-02-14
@license: GNU General Public License
@version: $Revision: 1.23 $

The bookmark tree used to graphically represent a Bookmark database
(L{Format}).
"""

from wxPython.wx import *
from Gui.Browser import Browser
from Data.Node import Bookmark, Folder, Separator
from Data.Formats import Format, FormatObserver, XBEL
from Gui.Helper import ErrorDialog, uconv

class TreeObserver:
    """Abstract base class for receiving observer events.  The class
    functions are called from the class TreeSubject.

    @see: L{TreeSubject}
    @see: U{Observer Pattern<http://c2.com/cgi/wiki?ObserverPattern>}
    @since: 2004-02-17
    """
    def selected(self, node):
        """A node has been selected.

        @param node: Selected node
        @type node: L{Node}
        """
        pass # implemented in concrete class

    def activated(self, bookmark):
        """A bookmark has been activated, i.e. double-clicked.

        @param bookmark: Activated bookmark
        @type bookmark: L{Bookmark}
        """
        pass # implemented in concrete class

class TreeSubject:
    """Abstract class which notifies subscribers about changes in a
    node.

    @see: U{Observer Pattern<http://c2.com/cgi/wiki?ObserverPattern>}
    @since: 2004-02-17
    """
    def __init__(self):
        """Constructor."""
        self.__observers = []

    def subscribe(self, observer):
        """Add the observer to the list of observers so that it gets
        notified on changes.

        @param observer: Observer to add
        @type observer: L{TreeObserver}
        """
        if not observer in self.__observers:
            self.__observers.append(observer)

    def unsubscribe(self, observer):
        """Remove the observer from the list of observers. It won't
        receive any events anymore.

        @param observer: Observer to remove
        @type observer: L{TreeObserver}
        """
        try:
            self.__observers.remove(observer)
        except ValueError:
            pass

    def _notifySelected(self, node):
        """Notify all observers that a node has been selected.

        @param node: Selected node
        @type node: L{Node}
        """
        for observer in self.__observers:
            observer.selected(node)

    def _notifyActivated(self, bookmark):
        """Notify all observers that a bookmark has been activated.

        @param bookmark: Activated bookmark
        @type bookmark: L{Bookmark}
        """
        for observer in self.__observers:
            observer.activated(bookmark)

class Tree(wxTreeCtrl,FormatObserver,TreeSubject):
    """The bookmark tree used to graphically represent a Bookmark
    database (L{Format}).
    
    @since: 2004-02-14
    """
    def __init__(self, parent, windowId, database):
        """Constructor. Initialize and paint the tree.

        @param parent: Window to paint the tree on
        @type parent: L{wxWindow}
        @param windowId: ID of this window
        @type windowId: L{wxWindowID}
        @param database: Database to work on
        @type database: L{Format}
        @raise TypeError: if database is not an object of a subclass
                          of L{Format}.
        """        
        if database is not None and not issubclass(database.__class__, Format):
            raise TypeError(_("Parameter database is not a valid Database object."))

        wxTreeCtrl.__init__(self, parent, windowId,
                style = wxTR_HAS_BUTTONS | wxTR_EDIT_LABELS | \
                        wxTR_HAS_VARIABLE_ROW_HEIGHT | wxTR_MULTIPLE)
        TreeSubject.__init__(self)
        self.__db = None
        self.setDatabase(database)

        # drag and drop
        EVT_TREE_BEGIN_DRAG(self, windowId, self.OnBeginDrag)
        EVT_TREE_END_DRAG(self, windowId, self.OnEndDrag)

        EVT_TREE_BEGIN_LABEL_EDIT(self, windowId, self.OnTreeLabelEditBegin)
        EVT_TREE_END_LABEL_EDIT(self, windowId, self.OnTreeLabelEditEnd)
        EVT_TREE_ITEM_ACTIVATED(self, windowId, self.OnTreeItemActivated)
        EVT_TREE_SEL_CHANGED(self, windowId, self.OnSelectionChanged)

    def created(self, node):
        """Called by FormatSubject when a node was created.

        @param node: New node
        @type node: L{Node}
        """
        self.__addNode(node)
        self.__dirty = True

    def __changeTitle(self, item):
        """This is a L{walkTree} function. (Re)sets the text of the
        tree node to the title of the associated bookmark/folder.

        @param item: Item of which the title is to be reset
        @type item: wxTreeItemID"""
        try:
            title = self.GetItemData(item).GetData().getTitle()
        except:
            title = u'------' # for separator
        self.SetItemText(item, title)

    def changed(self, node):
        """Called by FormatSubject when a node was changed. Refresh
        the title of the nodes item in the tree list.

        @param node: Modified node
        @type node: L{Node}
        """
        
        self.walkTree(self.__changeTitle, node)
        self.__dirty = True

    def __moveNode(self, item):
        """This is a L{walkTree} function. Moves the item to the new parent.

        @param item: Item of which the title is to be reset
        @type item: wxTreeItemID"""
        node = self.GetItemData(item).GetData()
        if node:
            self.Delete(item)
            self.__addNode(node)

    def moved(self, node):
        """Called by FormatSubject when a node was moved.

        @param node: Moved node
        @type node: L{Node}
        """
        self.walkTree(self.__moveNode, node)
        self.__dirty = True

    def __deleteNode(self, item):
        """This is a L{walkTree} function. This deletes the node from
        the list.

        @param item: Item to delete
        @type item: wxTreeItemID"""
        self.Delete(item)

    def deleted(self, node):
        """Called by FormatSubject when a node was deleted.

        @param node: Deleted node
        @type node: L{Node}
        """
        self.walkTree(self.__deleteNode, node)
        self.__dirty = True

    def setDatabase(self, database):
        """Set the database to be represented by this tree.

        @param database: Database to work on
        @type database: L{Format}
        """
        if self.__db != None:
            self.__db.unsubscribe(self)
        self.__db = database
        self.DeleteAllItems()
        if database is not None:
            if database.getFirst() is not None:
                self.__addNode(database.getFirst())
                self.Expand(self.GetRootItem())
            database.subscribe(self)
        self.__dirty = False

    def getDatabase(self):
        """Return the database represented by this tree.

        @return: Opened database
        @rtype: L{Format}
        """
        return self.__db

    def isDirty(self):
        """Return true if the database has been modified since being
        opened.

        @return: Dirty value
        @rtype: Boolean
        """
        return self.__dirty

    def __getTreeItemId(self, item):
        """This is a L{walkTree} function. It is used by L{__addNode}
        to determine the wxTreeItemId of a node.

        @param item: Item that matched
        @type item: wxTreeItemID"""
        self.__tmpTreeItemId = item

    def __addNode(self, node, parent = None):
        """Add a node to the tree.

        @param node: Node to add
        @type node: L{Node}
        @param parent: Parent tree item of node. If this is None, the
                       function determines it automatically.
        @type parent: L{wxTreeItemId}
        """
        if node.getParent() == None:
            treeNode = self.AddRoot(node.getTitle(), -1, -1, \
                       wxTreeItemData(node))
            self._notifySelected(node)
        else:
            if parent is None:
                # determine parent
                self.__tmpTreeItemId = None
                self.walkTree(self.__getTreeItemId, node.getParent())
                parent = self.__tmpTreeItemId
                del self.__tmpTreeItemId
                
                if parent is None:
                    # parent node not in tree view, add below root
                    parent = self.GetRootItem()
                elif not parent.IsOk():
                    parent = self.GetRootItem()
            try:
                title = node.getTitle()
            except:
                title = u'------' # for separator
            treeNode = self.AppendItem(parent, title, -1, -1, \
                       wxTreeItemData(node))
        try:
            child = node.getFirstChild()
            while child != None:
                self.__addNode(child, treeNode)
                child = child.getNextSibling()
        except:
            # method getFirstChild() only exists for Folder object
            pass

    #
    # Tree events
    #
    def OnTreeLabelEditBegin(self, event):
        """Called when inline editing of a tree item is started.

        @param event: Event object
        @type event: L{wxTreeEvent}
        """
        node = self.GetItemData(event.GetItem()).GetData()
        try:
            title = node.getTitle()
        except:
            # node does not have title attribute
            event.Veto()

    def OnTreeLabelEditEnd(self, event):
        """Called when inline editing of a tree item is finished.

        @param event: Event object
        @type event: L{wxTreeEvent}
        """
        node = self.GetItemData(event.GetItem()).GetData()
        if event.GetLabel() != '':
            label = uconv(event.GetLabel())
            if node.getTitle() != label:
                node.setTitle(label)
                self.__db._notifyChanged(node)
            else:
                event.Veto()

    def OnTreeItemActivated(self, event):
        """Called when a tree item has been activated. For example by
        double-clicking.

        @param event: Event object
        @type event: L{wxTreeEvent}
        """
        activated = self.GetItemData(event.GetItem()).GetData()
        if activated.canExecute():
            self._notifyActivated(activated)

    def OnSelectionChanged(self, event):
        """Called when a tree item has been selected.

        @param event: Event object
        @type event: L{wxTreeEvent}
        """
        self._notifySelected(self.GetItemData(event.GetItem()).GetData())

    def OnBeginDrag(self, event):
        """Called when the left mouse button is used to initiate
        dragging tree nodes.

        @param event: Event object
        @type event: L{wxTreeEvent}
        """
        for item in self.GetSelections():
            if item == self.GetRootItem():
                # do not allow dragging of root
                return
        self.__itemsBeingDragged = self.GetSelections()
        event.Allow()

    def OnEndDrag(self, event):
        """Called when the dragged node is dropped somewhere.

        @param event: Event object
        @type event: L{wxTreeEvent}
        """
        source = self.__itemsBeingDragged
        if source is None:
            return
        destination = event.GetItem()
        dstNode = self.GetItemData(destination).GetData()
        self.__itemsBeingDragged = None

        if destination.IsOk() and dstNode.canContainNodes():
            # check for all source nodes if they be dragged into selected
            # destination and do so if legal
            for srcItem in source:
                srcNode = self.GetItemData(srcItem).GetData()
                if srcItem != destination and \
                       not srcNode.containsDescendant(dstNode):
                    self.__db.move(srcNode, dstNode)

    #
    # helper functions
    #
    def __walkTree(self, node, func, base):
        """Recursive function used to implement the function
        walkTree(). Works on a node and all its subnodes.

        @return: True if the node base was found
        @rtype: Boolean
        @param node: Node to apply function to
        @type node: L{wxTreeItemId}
        @param func: The function to call for every node.
        @type func: A function accepting a wxTreeItemId param.
        @param base: Only call the function if the associated item
                     data is equal to this. If base is None, the
                     function is called for every node in the tree.
        @type base: L{Node}
        """
        if base == None:
            func(node)
        elif self.GetItemData(node).GetData() == base:
            
            func(node)
            return true

        child, cookie = self.GetFirstChild(node, 0)
        while child.IsOk():
            if self.__walkTree(child, func, base):
                return true
            child, cookie = self.GetNextChild(node, cookie)
        return false
        
    def walkTree(self, func, base = None):
        """Apply a function to each node in the tree.

        @param func: The function to call for every node.
        @type func: A function accepting a wxTreeItemId param.
        @param base: Only call the function if the associated item
                     data is equal to this.
        @type base: L{Node}
        """
        self.__walkTree(self.GetRootItem(), func, base)

    #
    # Selection functions
    #
    def getSelectedNode(self):
        """Returns the currently selected node. In a case where
        multiple items are selected, only the item which was selected
        last is returned.

        @return: Selected node
        @rtype: L{Node}
        """
        selected = self.GetSelections()
        if len(selected) > 0:
            return self.GetItemData(selected[-1]).GetData()
        else:
            return self.GetItemData(self.GetRootItem()).GetData()

    def getSelectedBookmark(self):
        """Returns the currently selected bookmark or None if any
        other node type is selected.

        @return: Selected bookmark or none
        @rtype: L{Bookmark}
        """
        selected = self.getSelectedNode()
        if selected.canExecute():
            return selected
        else:
            return None

    def getSelectedFolder(self):
        """Returns the currently selected folder. If a bookmark is
        selected, the folder's parent is returned.

        @return: Current folder
        @rtype: L{Folder}
        """
        selected = self.getSelectedNode()
        if selected.canContainNodes():
            return selected
        else:
            return selected.getParent()

    #
    # clipboard handling
    #
    def CopyNode(self):
        """Copy the currently selected node into the clipboard. It
        uses two formats. Our custom XBEL format (using L{wxNodeData}
        class) and normal text which the XBEL source.
        """
        if len(self.GetSelections()) == 0:
            return
        if wxTheClipboard.Open():
            nodeData = wxDataObjectComposite()
            if len(self.GetSelections()) > 1:
                import copy
                xbel = ""
                for item in self.GetSelections():
                    node = self.GetItemData(item).GetData()
                    if item == self.GetRootItem():
                        # can not copy root and subitems together, only return root
                        xbel = node.getXbel()
                        break
                    xbel += node.getXbel()
            else:
                xbel = self.getSelectedNode().getXbel()
            if xbel.strip().find("<xbel") != 0:
                xbel = "<xbel version=\"1.0\">" + xbel + "</xbel>"
            nodeData.Add(wxNodeData(xbel), True)
            nodeData.Add(wxTextDataObject(xbel))
            if not wxTheClipboard.SetData(nodeData):
                ErrorDialog(_("Could not copy data to clipboard."))
            wxTheClipboard.Close()
        else:
            ErrorDialog(_("Could not open clipboard."))

    def PasteNode(self):
        """Paste data from the clipboard. This uses the custom XBEL
        format. No other format is supported at the moment.
        """
        if not wxTheClipboard.IsSupported(wxCustomDataFormat("XBEL")):
            return

        if wxTheClipboard.Open():
            data = wxNodeData()
            if not wxTheClipboard.GetData(data):
                ErrorDialog(_("Can not access clipboard."))
                wxTheClipboard.Close()
                return

            xbel = XBEL()
            xbel.parse(data.GetData())
            node = xbel.getFirst().getFirstChild()
            while node is not None:
                if node.getParent() != xbel.getFirst():
                    # This check is necessary because we move the nodes around.
                    # The "next node" reference changes when moving the node.
                    break
                next = node.getNextSibling()
                node.setParent(self.getSelectedFolder())
                self.__db.add(node)
                node = next
            self.Expand(self.GetSelections()[0])
            wxTheClipboard.Close()
        else:
            ErrorDialog(_("Could not open clipboard."))

    def CutNode(self):
        """Copy selected node to the clipboard and delete the item
        after this. This method does not have any intelligence of its
        own, it calls the two functions L{CopyNode} and
        L{DeleteSelected}.
        """
        if self.GetRootItem() in self.GetSelections():
            ErrorDialog(_("You can not cut the root folder."))
            return
        self.CopyNode()
        self.DeleteSelected()

    #
    # Database functions
    #
    def NewBookmark(self):
        """Create a new bookmark in the currently selected folder."""
        self.__db.add(Bookmark(self.getSelectedFolder(), u'', \
                               u'', uconv(_('New bookmark'))))
        selection = self.GetSelections()[-1]
        new = self.GetLastChild(selection)
        if new.IsOk():
            self.SelectItem(new)
        else:
            new = self.GetLastChild(self.GetItemParent(selection))
            if new.IsOk():
                self.SelectItem(new)

    def NewFolder(self):
        """Create a new bookmark in the currently selected folder."""
        self.__db.add(Folder(self.getSelectedFolder(), u'', uconv(_('New folder'))))
        selection = self.GetSelections()[-1]
        new = self.GetLastChild(selection)
        if new.IsOk():
            self.SelectItem(new)
        else:
            new = self.GetLastChild(self.GetItemParent(selection))
            if new.IsOk():
                self.SelectItem(new)

    def NewSeparator(self):
        """Create a new separator in the currently selected folder."""
        self.__db.add(Separator(self.getSelectedFolder()))
        selection = self.GetSelections()[-1]
        new = self.GetLastChild(selection)
        if new.IsOk():
            self.SelectItem(new)
        else:
            new = self.GetLastChild(self.GetItemParent(selection))
            if new.IsOk():
                self.SelectItem(new)

    def DeleteSelected(self):
        """Delete the selected node.
        """
        for item in self.GetSelections():
            if item == self.GetRootItem():
                ErrorDialog(_("You can not delete the root folder."))
            else:
                node = self.GetItemData(item).GetData()
                self.__db.remove(node)

class MainTreeHandler(TreeObserver):
    """This is the main tree handler used for most tree objects. It
    implements all functions necessary for the user to be able to work
    with all trees in a similar way."""
    def __init__(self):
        """Initialize the class."""
        self.__browser = Browser()
    
    def activated(self, bookmark):
        """Open the activated bookmark in the default browser.

        @param bookmark: Activated bookmark
        @type bookmark: L{Bookmark}
        """
        if bookmark.getHref() != '':
            self.__browser.launch(bookmark.getHref())

class wxNodeData(wxCustomDataObject):
    """This class implements data storage for nodes. This is used by
    wxPython for copy&paste and drag&drop.

    @since: 2004-04-02
    """
    def __init__(self, data = ""):
        """Class constructor.

        @param data: XBEL data to store
        @type data: String
        """
        wxCustomDataObject.__init__(self, wxCustomDataFormat("XBEL"))
        self.SetData(data)
